# -*- coding: utf-8 -*-

from btb_benchmark.main import run_benchmark

__author__ = """MIT Data To AI Lab"""
__email__ = 'dailabmit@gmail.com'
__version__ = '0.4.1.dev0'
__all__ = ('run_benchmark', )
