# -*- coding: utf-8 -*-

"""Top level where all the metamodels are imported."""

from baytune.tuning.metamodels.gaussian_process import GaussianProcessMetaModel

__all__ = ("GaussianProcessMetaModel",)
